@@ -3,8 +3,6 @@
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
- * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
- * 
  * This file contains Original Code and/or Modifications of Original Code
  * as defined in and that are subject to the Apple Public Source License
  * Version 2.0 (the 'License'). You may not use this file except in
@@ -246,6 +244,53 @@ enum bool launching_with_prebound_libraries)
 	clear_being_linked_list(FALSE);
 }
 
+/*
+ * resolve_external_relocations_in_object_image() preforms the external
+ * relocations for just the one specified object image.
+ */
+void
+resolve_external_relocations_in_object_image(
+struct object_image *object_image)
+{
+    enum link_state link_state;
+    struct segment_command *linkedit_segment;
+    struct symtab_command *st;
+    struct dysymtab_command *dyst;
+    struct relocation_info *relocs;
+    struct nlist *symbols;
+    char *strings;
+
+	linkedit_segment = object_image->image.linkedit_segment;
+	st = object_image->image.st;
+	dyst = object_image->image.dyst;
+	relocs = (struct relocation_info *)
+	    (object_image->image.vmaddr_slide +
+		linkedit_segment->vmaddr +
+		dyst->extreloff -
+		linkedit_segment->fileoff);
+	symbols = (struct nlist *)
+	    (object_image->image.vmaddr_slide +
+		linkedit_segment->vmaddr +
+		st->symoff -
+		linkedit_segment->fileoff);
+	strings = (char *)
+	    (object_image->image.vmaddr_slide +
+		linkedit_segment->vmaddr +
+		st->stroff -
+		linkedit_segment->fileoff);
+
+	link_state = 
+	    external_relocation(
+		&(object_image->image),
+		relocs,
+		dyst->nextrel,
+		symbols,
+		strings,
+		NULL, /* library_name */
+		object_image->image.name);
+	SET_LINK_STATE(object_image->module, link_state);
+}
+
 /*
  * undo_prebound_lazy_pointers() undoes the prebinding for lazy symbol pointers.
  * This is also done by local_relocation() which is called if the image can't
